# -*- coding: utf-8 -*-

from .utils import TileReader, convert_to_ecef
from .tile import Tile
from .feature_table import Feature
from .gltf import GlTF
from .pnts import Pnts
from .b3dm import B3dm
from .batch_table import BatchTable
from .wkb_utils import TriangleSoup

__version__ = '1.0.2'
__all__ = ['TileReader', 'convert_to_ecef', 'Tile', 'Feature', 'GlTF', 'Pnts',
           'B3dm', 'BatchTable', 'TriangleSoup']



# -*- coding: utf-8 -*-
import struct
import numpy as np

from .tile import Tile, TileHeader, TileBody, TileType
from .gltf import GlTF
from .batch_table import BatchTable


class B3dm(Tile):

    @staticmethod
    def from_glTF(gltf, bt=None):
        """
        Parameters
        ----------
        gltf : GlTF
            glTF object representing a set of objects

        bt : Batch Table (optional)
            BatchTable object containing per-feature metadata

        Returns
        -------
        tile : Tile
        """

        tb = B3dmBody()
        tb.glTF = gltf
        tb.batch_table = bt

        th = B3dmHeader()
        th.sync(tb)

        t = Tile()
        t.body = tb
        t.header = th

        return t

    @staticmethod
    def from_array(array):
        """
        Parameters
        ----------
        array : numpy.array

        Returns
        -------
        t : Tile
        """

        # build tile header
        h_arr = array[0:B3dmHeader.BYTELENGTH]
        h = B3dmHeader.from_array(h_arr)

        if h.tile_byte_length != len(array):
            raise RuntimeError("Invalid byte length in header")

        # build tile body
        b_arr = (array[B3dmHeader.BYTELENGTH:h.tile_byte_length
                 - B3dmHeader.BYTELENGTH])
        b = B3dmBody.from_array(h, b_arr)

        # build Tile with header and body
        t = Tile()
        t.header = h
        t.body = b

        return t


class B3dmHeader(TileHeader):
    BYTELENGTH = 28

    def __init__(self):
        self.type = TileType.BATCHED3DMODEL
        self.magic_value = "b3dm"
        self.version = 1
        self.tile_byte_length = 0
        self.ft_json_byte_length = 0
        self.ft_bin_byte_length = 0
        self.bt_json_byte_length = 0
        self.bt_bin_byte_length = 0
        self.bt_length = 0  # number of models in the batch

    def to_array(self):
        header_arr = np.fromstring(self.magic_value, np.uint8)

        header_arr2 = np.array([self.version,
                                self.tile_byte_length,
                                self.ft_json_byte_length,
                                self.ft_bin_byte_length,
                                self.bt_json_byte_length,
                                self.bt_bin_byte_length], dtype=np.uint32)

        return np.concatenate((header_arr, header_arr2.view(np.uint8)))

    def sync(self, body):
        """
        Allow to synchronize headers with contents.
        """

        # extract array
        glTF_arr = body.glTF.to_array()

        # sync the tile header with feature table contents
        self.tile_byte_length = len(glTF_arr) + B3dmHeader.BYTELENGTH
        self.bt_json_byte_length = 0
        self.bt_bin_byte_length = 0
        self.ft_json_byte_length = 0
        self.ft_bin_byte_length = 0

        if body.batch_table is not None:
            bth_arr = body.batch_table.to_array()
            # btb_arr = body.batch_table.body.to_array()

            self.tile_byte_length += len(bth_arr)
            self.bt_json_byte_length = len(bth_arr)

        # fth_arr = body.feature_table.header.to_array()
        # ftb_arr = body.feature_table.body.to_array()

    @staticmethod
    def from_array(array):
        """
        Parameters
        ----------
        array : numpy.array

        Returns
        -------
        h : TileHeader
        """

        h = B3dmHeader()

        if len(array) != B3dmHeader.BYTELENGTH:
            raise RuntimeError("Invalid header length")

        h.magic_value = "b3dm"
        h.version = struct.unpack("i", array[4:8])[0]
        h.tile_byte_length = struct.unpack("i", array[8:12])[0]
        h.ft_json_byte_length = struct.unpack("i", array[12:16])[0]
        h.ft_bin_byte_length = struct.unpack("i", array[16:20])[0]
        h.bt_json_byte_length = struct.unpack("i", array[20:24])[0]
        h.bt_bin_byte_length = struct.unpack("i", array[24:28])[0]

        h.type = TileType.BATCHED3DMODEL

        return h


class B3dmBody(TileBody):
    def __init__(self):
        self.batch_table = BatchTable()
        # self.feature_table = FeatureTable()
        self.glTF = GlTF()

    def to_array(self):
        # TODO : export feature table
        array = self.glTF.to_array()
        if self.batch_table is not None:
            array = np.concatenate((self.batch_table.to_array(), array))
        return array

    @staticmethod
    def from_glTF(glTF):
        """
        Parameters
        ----------
        th : TileHeader

        glTF : GlTF

        Returns
        -------
        b : TileBody
        """

        # build tile body
        b = B3dmBody()
        b.glTF = glTF

        return b

    @staticmethod
    def from_array(th, array):
        """
        Parameters
        ----------
        th : TileHeader

        array : numpy.array

        Returns
        -------
        b : TileBody
        """

        # build feature table
        ft_len = th.ft_json_byte_length + th.ft_bin_byte_length
        # ft_arr = array[0:ft_len]
        # ft = FeatureTable.from_array(th, ft_arr)

        # build batch table
        bt_len = th.bt_json_byte_length + th.bt_bin_byte_length
        # bt_arr = array[ft_len:ft_len+bt_len]
        # bt = BatchTable.from_array(th, bt_arr)

        # build glTF
        glTF_len = (th.tile_byte_length - ft_len - bt_len
                    - B3dmHeader.BYTELENGTH)
        glTF_arr = array[ft_len+bt_len:ft_len+bt_len+glTF_len]
        glTF = GlTF.from_array(glTF_arr)

        # build tile body with feature table
        b = B3dmBody()
        # b.feature_table = ft
        # b.batch_table = bt
        b.glTF = glTF

        return b



# -*- coding: utf-8 -*-

import numpy as np
from enum import Enum
from abc import ABC, abstractmethod


class Tile(ABC):

    def __init__(self):
        self.header = None
        self.body = None

    def to_array(self):
        self.sync()
        header_arr = self.header.to_array()
        body_arr = self.body.to_array()
        return np.concatenate((header_arr, body_arr))

    def to_hex_str(self):
        arr = self.to_array()
        return " ".join("{:02X}".format(x) for x in arr)

    def save_as(self, filename):
        tile_arr = self.to_array()
        with open(filename, 'bw') as f:
            f.write(bytes(tile_arr))

    def sync(self):
        """
        Allow to synchronize headers with contents.
        """

        self.header.sync(self.body)


class TileType(Enum):

    UNKNWON = 0
    POINTCLOUD = 1
    BATCHED3DMODEL = 2


class TileHeader(ABC):
    @abstractmethod
    def from_array(self, array):
        pass

    @abstractmethod
    def to_array(self):
        pass

    @abstractmethod
    def sync(self, body):
        pass


class TileBody(object):
    @abstractmethod
    def to_array(self):
        pass



# -*- coding: utf-8 -*-
import struct
import numpy as np
import json


class GlTF(object):

    def __init__(self):
        self.header = {}
        self.body = None

    def to_array(self):  # glb
        scene = json.dumps(self.header, separators=(',', ':'))

        # body must be 4-byte aligned
        scene += ' '*((4 - len(scene) % 4) % 4)

        padding = np.array([0 for i in range(0, (4 - len(self.body) % 4) % 4)],
                           dtype=np.uint8)

        length = 28 + len(self.body) + len(scene) + len(padding)
        binaryHeader = np.array([0x46546C67,  # "glTF" magic
                                 2,  # version
                                 length], dtype=np.uint32)
        jsonChunkHeader = np.array([len(scene),  # JSON chunck length
                                    0x4E4F534A], dtype=np.uint32)  # "JSON"

        binChunkHeader = np.array([len(self.body) + len(padding),
                                   # BIN chunck length
                                   0x004E4942], dtype=np.uint32)  # "BIN"

        return np.concatenate((binaryHeader.view(np.uint8),
                               jsonChunkHeader.view(np.uint8),
                               np.fromstring(scene, dtype=np.uint8),
                               binChunkHeader.view(np.uint8),
                               self.body,
                               padding))

    @staticmethod
    def from_array(array):
        """
        Parameters
        ----------
        array : numpy.array

        Returns
        -------
        glTF : GlTf
        """

        glTF = GlTF()

        if struct.unpack("4s", array[0:4])[0] != b"glTF":
            raise RuntimeError("Array does not contain a binary glTF")

        if struct.unpack("i", array[4:8])[0] != 1:
            raise RuntimeError("Unsupported glTF version")

        length = struct.unpack("i", array[8:12])[0]
        content_length = struct.unpack("i", array[12:16])[0]

        if struct.unpack("i", array[16:20])[0] != 0:
            raise RuntimeError("Unsupported binary glTF content type")

        header = struct.unpack(str(content_length) + "s",
                               array[20:20+content_length])[0]
        glTF.header = json.loads(header.decode("ascii"))
        glTF.body = array[20+content_length:length]

        return glTF

    @staticmethod
    def from_binary_arrays(arrays, transform, binary=True, batched=True,
                           uri=None, textureUri=None):
        """
        Parameters
        ----------
        arrays : array of dictionaries
            Each dictionary has the data for one geometry
            arrays['position']: binary array of vertex positions
            arrays['normal']: binary array of vertex normals
            arrays['uv']: binary array of vertex texture coordinates
                          (Not implemented yet)
            arrays['bbox']: geometry bounding box (numpy.array)

        transform : numpy.array
            World coordinates transformation flattend matrix

        Returns
        -------
        glTF : GlTF
        """

        glTF = GlTF()

        textured = 'uv' in arrays[0]
        binVertices = []
        binNormals = []
        binIds = []
        binUvs = []
        nVertices = []
        bb = []
        batchLength = 0
        for i, geometry in enumerate(arrays):
            binVertices.append(geometry['position'])
            binNormals.append(geometry['normal'])
            n = round(len(geometry['position']) / 12)
            nVertices.append(n)
            bb.append(geometry['bbox'])
            if batched:
                binIds.append(np.full(n, i, dtype=np.float32))
            if textured:
                binUvs.append(geometry['uv'])

        if batched:
            binVertices = [b''.join(binVertices)]
            binNormals = [b''.join(binNormals)]
            binUvs = [b''.join(binUvs)]
            binIds = [b''.join(binIds)]
            nVertices = [sum(nVertices)]
            batchLength = len(arrays)
            [minx, miny, minz] = bb[0][0]
            [maxx, maxy, maxz] = bb[0][1]
            for box in bb[1:]:
                minx = min(minx, box[0][0])
                miny = min(miny, box[0][1])
                minz = min(minz, box[0][2])
                maxx = max(maxx, box[1][0])
                maxy = max(maxy, box[1][1])
                maxz = max(maxz, box[1][2])
            bb = [[[minx, miny, minz], [maxx, maxy, maxz]]]

        glTF.header = compute_header(binVertices, nVertices, bb, transform,
                                     textured, batched, batchLength, uri,
                                     textureUri)
        glTF.body = np.frombuffer(compute_binary(binVertices, binNormals,
                                  binIds, binUvs), dtype=np.uint8)

        return glTF


def compute_binary(binVertices, binNormals, binIds, binUvs):
    bv = b''.join(binVertices)
    bn = b''.join(binNormals)
    bid = b''.join(binIds)
    buv = b''.join(binUvs)
    return bv + bn + buv + bid


def compute_header(binVertices, nVertices, bb, transform,
                   textured, batched, batchLength, uri, textureUri):
    # Buffer
    meshNb = len(binVertices)
    sizeVce = []
    for i in range(0, meshNb):
        sizeVce.append(len(binVertices[i]))

    byteLength = 2 * sum(sizeVce)
    if textured:
        byteLength += int(round(2 * sum(sizeVce) / 3))
    if batched:
        byteLength += int(round(sum(sizeVce) / 3))
    buffers = [{
        'byteLength': byteLength
    }]
    if uri is not None:
        buffers["binary_glTF"]["uri"] = uri

    # Buffer view
    bufferViews = []
    # vertices
    bufferViews.append({
        'buffer': 0,
        'byteLength': sum(sizeVce),
        'byteOffset': 0,
        'target': 34962
    })
    bufferViews.append({
        'buffer': 0,
        'byteLength': sum(sizeVce),
        'byteOffset': sum(sizeVce),
        'target': 34962
    })
    if textured:
        bufferViews.append({
            'buffer': 0,
            'byteLength': int(round(2 * sum(sizeVce) / 3)),
            'byteOffset': 2 * sum(sizeVce),
            'target': 34962
        })
    if batched:
        bufferViews.append({
            'buffer': 0,
            'byteLength': int(round(sum(sizeVce) / 3)),
            'byteOffset': int(round(8 / 3 * sum(sizeVce))) if textured
            else 2 * sum(sizeVce),
            'target': 34962
        })

    # Accessor
    accessors = []
    for i in range(0, meshNb):
        # vertices
        accessors.append({
            'bufferView': 0,
            'byteOffset': sum(sizeVce[0:i]),
            'componentType': 5126,
            'count': nVertices[i],
            'max': [bb[i][0][1], bb[i][0][2], bb[i][0][0]],
            'min': [bb[i][1][1], bb[i][1][2], bb[i][1][0]],
            'type': "VEC3"
        })
        # normals
        accessors.append({
            'bufferView': 1,
            'byteOffset': sum(sizeVce[0:i]),
            'componentType': 5126,
            'count': nVertices[i],
            'max': [1, 1, 1],
            'min': [-1, -1, -1],
            'type': "VEC3"
        })
        if textured:
            accessors.append({
                'bufferView': 2,
                'byteOffset': int(round(2 / 3 * sum(sizeVce[0:i]))),
                'componentType': 5126,
                'count': sum(nVertices),
                'max': [1, 1],
                'min': [0, 0],
                'type': "VEC2"
            })
    if batched:
        accessors.append({
            'bufferView': 3 if textured else 2,
            'byteOffset': 0,
            'componentType': 5126,
            'count': nVertices[0],
            'max': [batchLength],
            'min': [0],
            'type': "SCALAR"
        })

    # Meshes
    meshes = []
    nAttributes = 3 if textured else 2
    for i in range(0, meshNb):
        meshes.append({
            'primitives': [{
                'attributes': {
                    "POSITION": nAttributes * i,
                    "NORMAL": nAttributes * i + 1
                },
                "material": 0,
                "mode": 4
            }]
        })
        if textured:
            meshes[i]['primitives'][0]['attributes']['TEXCOORD_0'] = (
                nAttributes * i + 2)
    if batched:
        meshes[0]['primitives'][0]['attributes']['_BATCHID'] = nAttributes

    # Nodes
    nodes = []
    for i in range(0, meshNb):
        nodes.append({
            'matrix': [float(e) for e in transform],
            'mesh': i
        })

    # Materials
    materials = [{
        'pbrMetallicRoughness': {
            'metallicFactor': 0
        },
        'name': 'Material',
    }]

    # Final glTF
    header = {
        'asset': {
            "generator": "py3dtiles",
            "version": "2.0"
        },
        'scene': 0,
        'scenes': [{
            'nodes': [i for i in range(0, len(nodes))]
        }],
        'nodes': nodes,
        'meshes': meshes,
        'materials': materials,
        'accessors': accessors,
        'bufferViews': bufferViews,
        'buffers': buffers
    }

    # Texture data
    if textured:
        header['textures'] = [{
            'sampler': 0,
            'source': 0
        }]
        header['images'] = [{
            'uri': textureUri
        }]
        header['samplers'] = [{
            "magFilter": 9729,
            "minFilter": 9987,
            "wrapS": 10497,
            "wrapT": 10497
        }]
        header['materials'][0]['pbrMetallicRoughness']['baseColorTexture'] = {
            'index': 0
        }

    return header



# -*- coding: utf-8 -*-

import json
from enum import Enum
import numpy as np


class Feature(object):

    def __init__(self):
        self.positions = {}
        self.colors = {}

    def to_array(self):
        pos_arr = np.array([(self.positions['X'], self.positions['Y'],
                            self.positions['Z'])]).view(np.uint8)[0]

        if len(self.colors):
            col_arr = np.array([(self.colors['Red'], self.colors['Green'],
                                self.colors['Blue'])]).view(np.uint8)[0]
        else:
            col_arr = np.array([])

        return [pos_arr, col_arr]

    @staticmethod
    def from_values(x, y, z, red=None, green=None, blue=None):
        f = Feature()

        f.positions = {'X': x, 'Y': y, 'Z': z}

        if red or green or blue:
            f.colors = {'Red': red, 'Green': green, 'Blue': blue}
        else:
            f.colors = {}

        return f

    @staticmethod
    def from_array(positions_dtype, positions, colors_dtype=None, colors=None):
        """
        Parameters
        ----------
        positions_dtype : numpy.dtype

        positions : numpy.array
            Array of uint8.

        colors_dtype : numpy.dtype

        colors : numpy.array
            Array of uint8.

        Returns
        -------
        f : Feature
        """

        f = Feature()

        # extract positions
        f.positions = {}
        off = 0
        for d in positions_dtype.names:
            dt = positions_dtype[d]
            data = np.array(positions[off:off+dt.itemsize]).view(dt)[0]
            off += dt.itemsize
            f.positions[d] = data

        # extract colors
        f.colors = {}
        if colors_dtype is not None:
            off = 0
            for d in colors_dtype.names:
                dt = colors_dtype[d]
                data = np.array(colors[off:off+dt.itemsize]).view(dt)[0]
                off += dt.itemsize
                f.colors[d] = data

        return f


class SemanticPoint(Enum):

    NONE = 0
    POSITION = 1
    POSITION_QUANTIZED = 2
    RGBA = 3
    RGB = 4
    RGB565 = 5
    NORMAL = 6
    NORMAL_OCT16P = 7
    BATCH_ID = 8


class FeatureTableHeader(object):

    def __init__(self):
        # point semantics
        self.positions = SemanticPoint.POSITION
        self.positions_offset = 0
        self.positions_dtype = None

        self.colors = SemanticPoint.NONE
        self.colors_offset = 0
        self.colors_dtype = None

        self.normal = SemanticPoint.NONE
        self.normal_offset = 0
        self.normal_dtype = None

        # global semantics
        self.points_length = 0
        self.rtc = None

    def to_array(self):
        jsond = self.to_json()
        json_str = json.dumps(jsond).replace(" ", "")
        n = len(json_str) + 28
        json_str += ' '*(4 - n % 4)
        return np.fromstring(json_str, dtype=np.uint8)

    def to_json(self):
        jsond = {}

        # length
        jsond['POINTS_LENGTH'] = self.points_length

        # rtc
        if self.rtc:
            jsond['RTC_CENTER'] = self.rtc

        # positions
        offset = {'byteOffset': self.positions_offset}
        if self.positions == SemanticPoint.POSITION:
            jsond['POSITION'] = offset
        elif self.positions == SemanticPoint.POSITION_QUANTIZED:
            jsond['POSITION_QUANTIZED'] = offset

        # colors
        offset = {'byteOffset': self.colors_offset}
        if self.colors == SemanticPoint.RGB:
            jsond['RGB'] = offset

        return jsond

    @staticmethod
    def from_dtype(positions_dtype, colors_dtype, npoints):
        """
        Parameters
        ----------
        positions_dtype : numpy.dtype
            Numpy description of a positions.

        colors_dtype : numpy.dtype
            Numpy description of a colors.

        Returns
        -------
        fth : FeatureTableHeader
        """

        fth = FeatureTableHeader()
        fth.points_length = npoints

        # search positions
        names = positions_dtype.names
        if ('X' in names) and ('Y' in names) and ('Z' in names):
            dtx = positions_dtype['X']
            dty = positions_dtype['Y']
            dtz = positions_dtype['Z']
            fth.positions_offset = 0
            if (dtx == np.float32 and dty == np.float32 and dtz == np.float32):
                fth.positions = SemanticPoint.POSITION
                fth.positions_dtype = np.dtype([('X', np.float32),
                                                ('Y', np.float32),
                                                ('Z', np.float32)])
            elif (dtx == np.uint16 and dty == np.uint16 and dtz == np.uint16):
                fth.positions = SemanticPoint.POSITION_QUANTIZED
                fth.positions_dtype = np.dtype([('X', np.uint16),
                                                ('Y', np.uint16),
                                                ('Z', np.uint16)])

        # search colors
        if colors_dtype is not None:
            names = colors_dtype.names
            if ('Red' in names) and ('Green' in names) and ('Blue' in names):
                if 'Alpha' in names:
                    fth.colors = SemanticPoint.RGBA
                    fth.colors_dtype = np.dtype([('Red', np.uint8),
                                                 ('Green', np.uint8),
                                                 ('Blue', np.uint8),
                                                 ('Alpha', np.uint8)])
                else:
                    fth.colors = SemanticPoint.RGB
                    fth.colors_dtype = np.dtype([('Red', np.uint8),
                                                 ('Green', np.uint8),
                                                 ('Blue', np.uint8)])

                fth.colors_offset = (fth.positions_offset
                                     + npoints*fth.positions_dtype.itemsize)
        else:
            fth.colors = SemanticPoint.NONE
            fth.colors_dtype = None

        return fth

    @staticmethod
    def from_array(array):
        """
        Parameters
        ----------
        array : numpy.array
            Json in 3D Tiles format. See py3dtiles/doc/semantics.json for an
            example.

        Returns
        -------
        fth : FeatureTableHeader
        """

        jsond = json.loads(array.tostring().decode('utf-8'))
        fth = FeatureTableHeader()

        # search position
        if "POSITION" in jsond:
            fth.positions = SemanticPoint.POSITION
            fth.positions_offset = jsond['POSITION']['byteOffset']
            fth.positions_dtype = np.dtype([('X', np.float32),
                                            ('Y', np.float32),
                                            ('Z', np.float32)])
        elif "POSITION_QUANTIZED" in jsond:
            fth.positions = SemanticPoint.POSITION_QUANTIZED
            fth.positions_offset = jsond['POSITION_QUANTIZED']['byteOffset']
            fth.positions_dtype = np.dtype([('X', np.uint16),
                                            ('Y', np.uint16),
                                            ('Z', np.uint16)])
        else:
            fth.positions = SemanticPoint.NONE
            fth.positions_offset = 0
            fth.positions_dtype = None

        # search colors
        if "RGB" in jsond:
            fth.colors = SemanticPoint.RGB
            fth.colors_offset = jsond['RGB']['byteOffset']
            fth.colors_dtype = np.dtype([('Red', np.uint8),
                                         ('Green', np.uint8),
                                         ('Blue', np.uint8)])
        else:
            fth.colors = SemanticPoint.NONE
            fth.colors_offset = 0
            fth.colors_dtype = None

        # points length
        if "POINTS_LENGTH" in jsond:
            fth.points_length = jsond["POINTS_LENGTH"]

        # RTC (Relative To Center)
        if "RTC_CENTER" in jsond:
            fth.rtc = jsond['RTC_CENTER']
        else:
            fth.rtc = None

        return fth


class FeatureTableBody(object):

    def __init__(self):
        self.positions_arr = []
        self.positions_itemsize = 0

        self.colors_arr = []
        self.colors_itemsize = 0

    def to_array(self):
        arr = self.positions_arr
        if len(self.colors_arr):
            arr = np.concatenate((self.positions_arr, self.colors_arr))
        return arr

    @staticmethod
    def from_features(fth, features):

        b = FeatureTableBody()

        # extract positions
        b.positions_itemsize = fth.positions_dtype.itemsize
        b.positions_arr = np.array([], dtype=np.uint8)

        if fth.colors_dtype is not None:
            b.colors_itemsize = fth.colors_dtype.itemsize
            b.colors_arr = np.array([], dtype=np.uint8)

        for f in features:
            fpos, fcol = f.to_array()
            b.positions_arr = np.concatenate((b.positions_arr, fpos))
            if fth.colors_dtype is not None:
                b.colors_arr = np.concatenate((b.colors_arr, fcol))

        return b

    @staticmethod
    def from_array(fth, array):
        """
        Parameters
        ----------
        header : FeatureTableHeader

        array : numpy.array

        Returns
        -------
        ftb : FeatureTableBody
        """

        b = FeatureTableBody()

        npoints = fth.points_length

        # extract positions
        pos_size = fth.positions_dtype.itemsize
        pos_offset = fth.positions_offset
        b.positions_arr = array[pos_offset:pos_offset+npoints*pos_size]
        b.positions_itemsize = pos_size

        # extract colors
        if fth.colors != SemanticPoint.NONE:
            col_size = fth.colors_dtype.itemsize
            col_offset = fth.colors_offset
            b.colors_arr = array[col_offset:col_offset+col_size*npoints]
            b.colors_itemsize = col_size

        return b

    def positions(self, n):
        itemsize = self.positions_itemsize
        return self.positions_arr[n*itemsize:(n+1)*itemsize]

    def colors(self, n):
        if len(self.colors_arr):
            itemsize = self.colors_itemsize
            return self.colors_arr[n*itemsize:(n+1)*itemsize]
        return []


class FeatureTable(object):

    def __init__(self):
        self.header = FeatureTableHeader()
        self.body = FeatureTableBody()

    def npoints(self):
        return self.header.points_length

    def to_array(self):
        fth_arr = self.header.to_array()
        ftb_arr = self.body.to_array()
        return np.concatenate((fth_arr, ftb_arr))

    @staticmethod
    def from_array(th, array):
        """
        Parameters
        ----------
        th : TileHeader

        array : numpy.array

        Returns
        -------
        ft : FeatureTable
        """

        # build feature table header
        fth_len = th.ft_json_byte_length
        fth_arr = array[0:fth_len]
        fth = FeatureTableHeader.from_array(fth_arr)

        # build feature table body
        ftb_len = th.ft_bin_byte_length
        ftb_arr = array[fth_len:fth_len+ftb_len]
        ftb = FeatureTableBody.from_array(fth, ftb_arr)

        # build feature table
        ft = FeatureTable()
        ft.header = fth
        ft.body = ftb

        return ft

    @staticmethod
    def from_features(pdtype, cdtype, features):
        """
        pdtype : numpy.dtype
            Numpy description for positions.

        cdtype : numpy.dtype
            Numpy description for colors.

        features : Feature[]

        Returns
        -------
        ft : FeatureTable
        """

        fth = FeatureTableHeader.from_dtype(pdtype, cdtype, len(features))
        ftb = FeatureTableBody.from_features(fth, features)

        ft = FeatureTable()
        ft.header = fth
        ft.body = ftb

        return ft

    def feature(self, n):
        pos = self.body.positions(n)
        col = self.body.colors(n)
        return Feature.from_array(self.header.positions_dtype, pos,
                                  self.header.colors_dtype, col)



# -*- coding: utf-8 -*-

import numpy as np
import json


class BatchTable(object):
    """
    Only the JSON header has been implemented for now. According to the batch
    table documentation, the binary body is useful for storing long arrays of
    data (better performances)
    """

    def __init__(self):
        self.header = {}

    def add_property_from_array(self, propertyName, array):
        self.header[propertyName] = array

    # returns batch table as binary
    def to_array(self):
        # convert dict to json string
        bt_json = json.dumps(self.header, separators=(',', ':'))
        # header must be 4-byte aligned (refer to batch table documentation)
        bt_json += ' '*(4 - len(bt_json) % 4)
        # returns an array of binaries representing the batch table
        return np.fromstring(bt_json, dtype=np.uint8)



# -*- coding: utf-8 -*-

import unittest
import numpy as np
# np.set_printoptions(formatter={'int':hex})

from py3dtiles import TileReader, Tile, Feature, Pnts


class TestTileReader(unittest.TestCase):

    def test_read(self):
        tile = TileReader().read_file('tests/pointCloudRGB.pnts')

        self.assertEqual(tile.header.version, 1.0)
        self.assertEqual(tile.header.tile_byte_length, 15176)
        self.assertEqual(tile.header.ft_json_byte_length, 148)
        self.assertEqual(tile.header.ft_bin_byte_length, 15000)
        self.assertEqual(tile.header.bt_json_byte_length, 0)
        self.assertEqual(tile.header.bt_bin_byte_length, 0)

        feature_table = tile.body.feature_table
        feature = feature_table.feature(0)
        dcol_res = {'Red': 44, 'Blue': 209, 'Green': 243}
        self.assertDictEqual(dcol_res, feature.colors)


class TestTileBuilder(unittest.TestCase):

    def test_build_without_colors(self):
        tread = TileReader().read_file('tests/pointCloudRGB.pnts')
        f0_ref = tread.body.feature_table.feature(0).positions

        # numpy dtype for positions and colors
        pdt = np.dtype([('X', '<f4'), ('Y', '<f4'), ('Z', '<f4')])

        # create features
        features = []
        for i in range(0, tread.body.feature_table.header.points_length):
            f = tread.body.feature_table.feature(i)
            p = f.positions
            pos = np.array([(p['X'], p['Y'], p['Z'])], dtype=pdt).view('uint8')
            newf = Feature.from_array(pdt, pos)
            features.append(newf)

        # create a tile
        t = Pnts.from_features(pdt, None, features)

        # configure the tile
        rtc = [1215012.8828876738, -4736313.051199594, 4081605.22126042]
        t.body.feature_table.header.rtc = rtc

        # get an array
        tile_arr = t.to_array()
        t2 = Pnts.from_array(tile_arr)
        self.assertEqual(t2.header.version, 1.0)
        self.assertEqual(t2.header.tile_byte_length, 12152)
        self.assertEqual(t2.header.ft_json_byte_length, 124)
        self.assertEqual(t2.header.ft_bin_byte_length, 12000)
        self.assertEqual(t2.header.bt_json_byte_length, 0)
        self.assertEqual(t2.header.bt_bin_byte_length, 0)

        feature_table = t.body.feature_table
        f0 = feature_table.feature(0).positions

        self.assertAlmostEqual(f0_ref['X'], f0['X'])
        self.assertAlmostEqual(f0_ref['Y'], f0['Y'])
        self.assertAlmostEqual(f0_ref['Z'], f0['Z'])

    def test_build(self):
        tread = TileReader().read_file('tests/pointCloudRGB.pnts')

        # numpy dtype for positions and colors
        pdt = np.dtype([('X', '<f4'), ('Y', '<f4'), ('Z', '<f4')])
        cdt = np.dtype([('Red', 'u1'), ('Green', 'u1'), ('Blue', 'u1')])

        # create features
        features = []
        for i in range(0, tread.body.feature_table.header.points_length):
            f = tread.body.feature_table.feature(i)
            p = f.positions
            c = f.colors
            pos = np.array([(p['X'], p['Y'], p['Z'])], dtype=pdt).view('uint8')
            col = np.array([(c['Red'], c['Green'], c['Blue'])],
                           dtype=cdt).view('uint8')
            newf = Feature.from_array(pdt, pos, cdt, col)
            features.append(newf)

        # create a tile
        t = Pnts.from_features(pdt, cdt, features)

        # configure the tile
        rtc = [1215012.8828876738, -4736313.051199594, 4081605.22126042]
        t.body.feature_table.header.rtc = rtc

        # get an array
        tile_arr = t.to_array()
        t2 = Pnts.from_array(tile_arr)
        self.assertEqual(t2.header.version, 1.0)
        self.assertEqual(t2.header.tile_byte_length, 15176)
        self.assertEqual(t2.header.ft_json_byte_length, 148)
        self.assertEqual(t2.header.ft_bin_byte_length, 15000)
        self.assertEqual(t2.header.bt_json_byte_length, 0)
        self.assertEqual(t2.header.bt_bin_byte_length, 0)

        feature_table = t.body.feature_table
        feature = feature_table.feature(0)
        dcol_res = {'Red': 44, 'Blue': 209, 'Green': 243}
        self.assertDictEqual(dcol_res, feature.colors)

        # t2.save_as("/tmp/py3dtiles_test_build_1.pnts")



import numpy as np
import math
import struct
from .earcut import earcut


class TriangleSoup:
    def __init__(self):
        self.triangles = []

    @staticmethod
    def from_wkb_multipolygon(wkb, associatedData=[]):
        """
        Parameters
        ----------
        wkb : string
            Well-Known Binary binary string describing a multipolygon

        associatedData : array
            array of multipolygons containing data attached to the wkb
            parameter multipolygon. Must be the same size as wkb.

        Returns
        -------
        ts : TriangleSoup
        """
        multipolygons = [parse(bytes(wkb))]

        for additionalWkb in associatedData:
            multipolygons.append(parse(bytes(additionalWkb)))

        trianglesArray = [[] for _ in range(len(multipolygons))]
        for i in range(0, len(multipolygons[0])):
            polygon = multipolygons[0][i]
            additionalPolygons = [mp[i] for mp in multipolygons[1:]]
            triangles = triangulate(polygon, additionalPolygons)
            for array, tri in zip(trianglesArray, triangles):
                array += tri
            """if(len(polygon) != 1):
                print("No support for inner polygon rings")
            else:
                if(len(polygon[0]) > 3):
                    triangles = triangulate(polygon[0],
                                            [p[0] for p in additionalPolygons])
                    for array, tri in zip(trianglesArray, triangles):
                        array += tri
                else:
                    for array, tri in zip(trianglesArray,
                                          [polygon] + additionalPolygons):
                        array += tri"""

        ts = TriangleSoup()
        ts.triangles = trianglesArray

        return ts

    def getPositionArray(self):
        """
        Parameters
        ----------

        Returns
        -------
        Binary array of vertice positions
        """

        verticeTriangles = self.triangles[0]
        verticeArray = vertexAttributeToArray(verticeTriangles)
        return b''.join(verticeArray)

    def getDataArray(self, index):
        """
        Parameters
        ----------
        index: int
            The index of the associated data

        Returns
        -------
        Binary array of vertice data
        """

        verticeTriangles = self.triangles[1 + index]
        verticeArray = vertexAttributeToArray(verticeTriangles)
        return b''.join(verticeArray)

    def getNormalArray(self):
        """
        Parameters
        ----------

        Returns
        -------
        Binary array of vertice normals
        """
        normals = []
        for t in self.triangles[0]:
            U = t[1] - t[0]
            V = t[2] - t[0]
            N = np.cross(U, V)
            norm = np.linalg.norm(N)
            if norm == 0:
                normals.append(np.array([0, 0, 1], dtype=np.float32))
            else:
                normals.append(N / norm)

        verticeArray = faceAttributeToArray(normals)
        return b''.join(verticeArray)

    def getBbox(self):
        """
        Parameters
        ----------

        Returns
        -------
        Array [[minX, minY, minZ],[maxX, maxY, maxZ]]
        """
        mins = np.array([np.min(t, 0) for t in self.triangles[0]])
        maxs = np.array([np.max(t, 0) for t in self.triangles[0]])
        return [np.min(mins, 0), np.max(maxs, 0)]


def faceAttributeToArray(triangles):
    array = []
    for face in triangles:
        array += [face, face, face]
    return array


def vertexAttributeToArray(triangles):
    array = []
    for face in triangles:
        for vertex in face:
            array.append(vertex)
    return array


def parse(wkb):
    multipolygon = []
    # length = len(wkb)
    # print(length)
    byteorder = struct.unpack('b', wkb[0:1])
    bo = '<' if byteorder[0] else '>'
    geomtype = struct.unpack(bo + 'I', wkb[1:5])[0]
    hasZ = (geomtype == 1006) or (geomtype == 1015)
    # MultipolygonZ or polyhedralSurface
    pntOffset = 24 if hasZ else 16
    pntUnpack = 'ddd' if hasZ else 'dd'
    geomNb = struct.unpack(bo + 'I', wkb[5:9])[0]
    # print(struct.unpack('b', wkb[9:10])[0])
    # print(struct.unpack('I', wkb[10:14])[0])   # 1003 (Polygon)
    # print(struct.unpack('I', wkb[14:18])[0])   # num lines
    # print(struct.unpack('I', wkb[18:22])[0])   # num points
    offset = 9
    for i in range(0, geomNb):
        offset += 5  # struct.unpack('bI', wkb[offset:offset+5])[0]
        # 1 (byteorder), 1003 (Polygon)
        lineNb = struct.unpack(bo + 'I', wkb[offset:offset+4])[0]
        offset += 4
        polygon = []
        for j in range(0, lineNb):
            pointNb = struct.unpack(bo + 'I', wkb[offset:offset+4])[0]
            offset += 4
            line = []
            for k in range(0, pointNb-1):
                pt = np.array(struct.unpack(bo + pntUnpack, wkb[offset:offset
                              + pntOffset]), dtype=np.float32)
                offset += pntOffset
                line.append(pt)
            offset += pntOffset   # skip redundant point
            polygon.append(line)
        multipolygon.append(polygon)
    return multipolygon


def triangulate(polygon, additionalPolygons=[]):
    """
    Triangulates 3D polygons
    """
    vect1 = polygon[0][1] - polygon[0][0]
    vect2 = polygon[0][2] - polygon[0][0]
    vectProd = np.cross(vect1, vect2)
    polygon2D = []
    holes = []
    delta = 0
    for p in polygon[:-1]:
        holes.append(delta + len(p))
        delta += len(p)
    # triangulation of the polygon projected on planes (xy) (zx) or (yz)
    if(math.fabs(vectProd[0]) > math.fabs(vectProd[1])
       and math.fabs(vectProd[0]) > math.fabs(vectProd[2])):
        # (yz) projection
        for linestring in polygon:
            for point in linestring:
                polygon2D.extend([point[1], point[2]])
    elif(math.fabs(vectProd[1]) > math.fabs(vectProd[2])):
        # (zx) projection
        for linestring in polygon:
            for point in linestring:
                polygon2D.extend([point[0], point[2]])
    else:
        # (xy) projextion
        for linestring in polygon:
            for point in linestring:
                polygon2D.extend([point[0], point[1]])

    trianglesIdx = earcut(polygon2D, holes, 2)

    arrays = [[] for _ in range(len(additionalPolygons) + 1)]
    for i in range(0, len(trianglesIdx), 3):
        t = trianglesIdx[i:i+3]
        p0 = unflatten(polygon, holes, t[0])
        p1 = unflatten(polygon, holes, t[1])
        p2 = unflatten(polygon, holes, t[2])
        # triangulation may break triangle orientation, test it before
        # adding triangles
        crossProduct = np.cross(p1 - p0, p2 - p0)
        invert = np.dot(vectProd, crossProduct) < 0
        if invert:
            arrays[0].append([p1, p0, p2])
        else:
            arrays[0].append([p0, p1, p2])
        for array, p in zip(arrays[1:],  additionalPolygons):
            pp0 = unflatten(p, holes, t[0])
            pp1 = unflatten(p, holes, t[1])
            pp2 = unflatten(p, holes, t[2])
            if invert:
                array.append([pp1, pp0, pp2])
            else:
                array.append([pp0, pp1, pp2])

    return arrays


def unflatten(array, lengths, index):
    for i in reversed(range(0, len(lengths))):
        lgth = lengths[i]
        if index >= lgth:
            return array[i + 1][index - lgth]
    return array[0][index]



# -*- coding: utf-8 -*-

import unittest

from py3dtiles import convert_to_ecef


class TestConvert(unittest.TestCase):

    def test_convert(self):

        # results tested with gdaltransform
        [x, y, z] = convert_to_ecef(-75.61200462622627,
                                    40.03886513981721,
                                    2848.448771114095,
                                    4326)
        self.assertAlmostEqual(x, 1215626.30684538)
        self.assertAlmostEqual(y, -4738673.45914053)
        self.assertAlmostEqual(z, 4083122.83975827)



# -*- coding: utf-8 -*-
import os
import re
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))


requirements = (
    'numpy',
    'pyproj',
    'cython',
    'triangle',
    'psycopg2-binary'
)

dev_requirements = (
    'pytest',
    'pytest-cov',
)

doc_requirements = (
    'sphinx',
    'sphinx_rtd_theme',
)

prod_requirements = (
)

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def find_version(*file_paths):
    """
    see https://github.com/pypa/sampleproject/blob/master/setup.py
    """

    with open(os.path.join(here, *file_paths), 'r') as f:
        version_file = f.read()

    # The version line must have the form
    # __version__ = 'ver'
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string. "
                       "Should be at the first line of __init__.py.")


setup(
    name='py3dtiles',
    version=find_version('py3dtiles', '__init__.py'),
    description="Python module for 3D tiles format",
    long_description=read('README.rst'),
    url='https://github.com/Oslandia/py3dtiles',
    author='dev',
    author_email='contact@oslandia.com',
    license='LGPL2 or later',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    packages=find_packages(),
    install_requires=requirements,
    test_suite="tests",
    scripts=["tools/py3dtiles_info", "tools/export_tileset"],
    extras_require={
        'dev': dev_requirements,
        'prod': prod_requirements,
        'doc': doc_requirements
    }
)



# -*- coding: utf-8 -*-

import unittest
import numpy as np
import binascii
import json
# np.set_printoptions(formatter={'int':hex})

from py3dtiles import TileReader, Tile, Feature, B3dm, GlTF, TriangleSoup


class TestTileReader(unittest.TestCase):

    def test_read(self):
        tile = TileReader().read_file('tests/dragon_low.b3dm')

        self.assertEqual(tile.header.version, 1.0)
        self.assertEqual(tile.header.tile_byte_length, 47246)
        self.assertEqual(tile.header.ft_json_byte_length, 20)
        self.assertEqual(tile.header.ft_bin_byte_length, 0)
        self.assertEqual(tile.header.bt_json_byte_length, 0)
        self.assertEqual(tile.header.bt_bin_byte_length, 0)

        with open('tests/dragon_low_gltf_header.json', 'r') as f:
            gltf_header = json.loads(f.read())
        self.assertDictEqual(gltf_header, tile.body.glTF.header)


class TestTileBuilder(unittest.TestCase):

    def test_build(self):
        with open('tests/building.wkb', 'rb') as f:
            wkb = f.read()
        ts = TriangleSoup.from_wkb_multipolygon(wkb)
        positions = ts.getPositionArray()
        normals = ts.getNormalArray()
        box = [[-8.74748499994166, -7.35523200035095, -2.05385796777344],
               [8.8036420000717, 7.29930999968201, 2.05386103222656]]
        arrays = [{
            'position': positions,
            'normal': normals,
            'bbox': box
        }]

        transform = np.array([
            [1, 0, 0, 1842015.125],
            [0, 1, 0, 5177109.25],
            [0, 0, 1, 247.87364196777344],
            [0, 0, 0, 1]], dtype=float)
        # translation : 1842015.125, 5177109.25, 247.87364196777344
        transform = transform.flatten('F')
        glTF = GlTF.from_binary_arrays(arrays, transform)
        t = B3dm.from_glTF(glTF)

        # get an array
        tile_arr = t.to_array()
        self.assertEqual(t.header.version, 1.0)
        self.assertEqual(t.header.tile_byte_length, 2952)
        self.assertEqual(t.header.ft_json_byte_length, 0)
        self.assertEqual(t.header.ft_bin_byte_length, 0)
        self.assertEqual(t.header.bt_json_byte_length, 0)
        self.assertEqual(t.header.bt_bin_byte_length, 0)

        # t.save_as("/tmp/py3dtiles_test_build_1.b3dm")

class TestTexturedTileBuilder(unittest.TestCase):

    def test_build(self):
        with open('tests/square.wkb', 'rb') as f:
            wkb = f.read()
        with open('tests/squareUV.wkb', 'rb') as f:
            wkbuv = f.read()
        ts = TriangleSoup.from_wkb_multipolygon(wkb, [wkbuv])
        positions = ts.getPositionArray()
        normals = ts.getNormalArray()
        uvs = ts.getDataArray(0)
        box = [[0, 0, 0],
               [10, 10, 0]]
        arrays = [{
            'position': positions,
            'normal': normals,
            'uv': uvs,
            'bbox': box
        }]

        transform = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1]], dtype=float)
        transform = transform.flatten('F')
        glTF = GlTF.from_binary_arrays(arrays, transform, textureUri='squaretexture.jpg')
        t = B3dm.from_glTF(glTF)

        # get an array
        tile_arr = t.to_array()
        self.assertEqual(t.header.version, 1.0)
        self.assertEqual(t.header.tile_byte_length, 1556)
        self.assertEqual(t.header.ft_json_byte_length, 0)
        self.assertEqual(t.header.ft_bin_byte_length, 0)
        self.assertEqual(t.header.bt_json_byte_length, 0)
        self.assertEqual(t.header.bt_bin_byte_length, 0)

        # t.save_as("/tmp/py3dtiles_test_build_1.b3dm")



# -*- coding: utf-8 -*-

import numpy as np
import pyproj
from .pnts import Pnts
from .b3dm import B3dm


def convert_to_ecef(x, y, z, epsg_input):
    inp = pyproj.Proj(init='epsg:{0}'.format(epsg_input))
    outp = pyproj.Proj(init='epsg:4978')  # ECEF
    return pyproj.transform(inp, outp, x, y, z)


class TileReader(object):

    def read_file(self, filename):
        with open(filename, 'rb') as f:
            data = f.read()
            arr = np.frombuffer(data, dtype=np.uint8)
            return self.read_array(arr)
        return None

    def read_array(self, array):
        magic = ''.join([c.decode('UTF-8') for c in array[0:4].view('c')])
        if magic == 'pnts':
            return Pnts.from_array(array)
        if magic == 'b3dm':
            return B3dm.from_array(array)
        return None



# -*- coding: utf-8 -*-
import struct
import numpy as np

from .tile import Tile, TileHeader, TileBody, TileType
from .feature_table import FeatureTable


class Pnts(Tile):

    @staticmethod
    def from_features(pdtype, cdtype, features):
        """
        Parameters
        ----------
        dtype : numpy.dtype
            Numpy description of a single feature

        features : Feature[]

        Returns
        -------
        tile : Tile
        """

        ft = FeatureTable.from_features(pdtype, cdtype, features)

        tb = PntsBody()
        tb.feature_table = ft

        th = PntsHeader()

        t = Tile()
        t.body = tb
        t.header = th

        return t

    @staticmethod
    def from_array(array):
        """
        Parameters
        ----------
        array : numpy.array

        Returns
        -------
        t : Tile
        """

        # build tile header
        h_arr = array[0:PntsHeader.BYTELENGTH]
        h = PntsHeader.from_array(h_arr)

        if h.tile_byte_length != len(array):
            raise RuntimeError("Invalid byte length in header")

        # build tile body
        b_len = h.ft_json_byte_length + h.ft_bin_byte_length
        b_arr = array[PntsHeader.BYTELENGTH:PntsHeader.BYTELENGTH+b_len]
        b = PntsBody.from_array(h, b_arr)

        # build Tile with header and body
        t = Tile()
        t.header = h
        t.body = b

        return t


class PntsHeader(TileHeader):
    BYTELENGTH = 28

    def __init__(self):
        self.type = TileType.POINTCLOUD
        self.magic_value = "pnts"
        self.version = 1
        self.tile_byte_length = 0
        self.ft_json_byte_length = 0
        self.ft_bin_byte_length = 0
        self.bt_json_byte_length = 0
        self.bt_bin_byte_length = 0

    def to_array(self):
        header_arr = np.fromstring(self.magic_value, np.uint8)

        header_arr2 = np.array([self.version,
                                self.tile_byte_length,
                                self.ft_json_byte_length,
                                self.ft_bin_byte_length,
                                self.bt_json_byte_length,
                                self.bt_bin_byte_length], dtype=np.uint32)

        return np.concatenate((header_arr, header_arr2.view(np.uint8)))

    def sync(self, body):
        """
        Allow to synchronize headers with contents.
        """

        # extract array
        fth_arr = body.feature_table.header.to_array()
        ftb_arr = body.feature_table.body.to_array()

        # sync the tile header with feature table contents
        self.tile_byte_length = (len(fth_arr) + len(ftb_arr) +
                                 PntsHeader.BYTELENGTH)
        self.ft_json_byte_length = len(fth_arr)
        self.ft_bin_byte_length = len(ftb_arr)

    @staticmethod
    def from_array(array):
        """
        Parameters
        ----------
        array : numpy.array

        Returns
        -------
        h : TileHeader
        """

        h = PntsHeader()

        if len(array) != PntsHeader.BYTELENGTH:
            raise RuntimeError("Invalid header length")

        h.magic_value = "pnts"
        h.version = struct.unpack("i", array[4:8])[0]
        h.tile_byte_length = struct.unpack("i", array[8:12])[0]
        h.ft_json_byte_length = struct.unpack("i", array[12:16])[0]
        h.ft_bin_byte_length = struct.unpack("i", array[16:20])[0]
        h.bt_json_byte_length = struct.unpack("i", array[20:24])[0]
        h.bt_bin_byte_length = struct.unpack("i", array[24:28])[0]

        h.type = TileType.POINTCLOUD

        return h


class PntsBody(TileBody):
    def __init__(self):
        self.feature_table = FeatureTable()
        # TODO : self.batch_table = BatchTable()

    def to_array(self):
        return self.feature_table.to_array()

    @staticmethod
    def from_array(th, array):
        """
        Parameters
        ----------
        th : TileHeader

        array : numpy.array

        Returns
        -------
        b : TileBody
        """

        # build feature table
        ft_len = th.ft_json_byte_length + th.ft_bin_byte_length
        ft_arr = array[0:ft_len]
        ft = FeatureTable.from_array(th, ft_arr)

        # build batch table
        # bt_len = th.bt_json_byte_length + th.bt_bin_byte_length
        # bt_arr = array[ft_len:ft_len+ba_len]
        # bt = BatchTable.from_array(th, bt_arr)

        # build tile body with feature table
        b = PntsBody()
        b.feature_table = ft
        # b.batch_table = bt

        return b
